/* Funny little demo framework, provides a number of useful functions
 * for 4k/64k coding.
 *
 * Urs of Mercury <urs@nerd2nerd.org>
 *
 * Includes code stolen from las <las@mercury-domination.de>
 */
/* Useful Defines you can apply:
 * DEBUG - well, duh.
 * MIPMAP - Do mipmapping on textures
 * IMG_LOAD - Load images via stb_image (otherwise: only BMP support through SDL)
 */
#ifdef LINUX
#include <SDL/SDL.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glext.h>
#else
#include <windows.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include "glext.h"
#include "glfn.h"
#endif
#include <math.h>
#ifdef DEBUG
#include <GL/glu.h>
#include <stdlib.h>
#include <stdio.h>
#endif /* DEBUG */

/* Yay, imageloading */
#ifdef IMG_LOAD
#ifndef DUMP_VIDEO
#define STBI_NO_WRITE
#endif
#define STBI_NO_FAILURE_STRINGS
#define STBI_NO_HDR
#include "stb_image.h"
#endif /* IMG_LOAD */

#ifdef LINUX
#define __small__ __attribute__((regparm(3),noinline))
#else
#define __small__
#endif

/* Data types for meshes exported by blender */
#ifndef BLENDER_EXPORT_MESH_DATA
#define BLENDER_EXPORT_MESH_DATA
typedef struct TexturedVertexData3D { float v[3],n[3],t[2]; } TexturedVertexData3D;
typedef struct ColoredVertexData3D { float v[3],n[3],c[4]; } ColoredVertexData3D;
typedef struct VertexData3D { float v[3],n[3]; } VertexData3D;
#endif /* BLENDER_EXPORT_MESH_DATA */

/******* Part 1: Arithmetic fun. ********/
/* Random Number generator */
static unsigned int random_state;
static unsigned int __small__ my_rnd(void) {

	const unsigned int random_const = 0x0f31782ce;

	random_state+=random_const+random_state>>3;
	random_state>>=1;
	random_state^=random_const+random_state<<3;
	//random_state>>=random_state & 1;

	return random_state;
}
/* Random between -1 and 1 */
static float __small__ my_float_rand(void) {
	return ((int)my_rnd()-(int)my_rnd())/438860800.f;
}

/******** Part 2: System Calls *********/
/* exit(0) */
#ifdef LINUX
inline void __attribute__((noreturn)) my_exit(void) {
	asm ( \
		"movl $1,%eax\n" \
		"xor %ebx,%ebx\n" \
		"int $0x80\n" \
	);
}
#endif

/******** Part 3: SDL and opengl setup ******/
#ifdef DEBUG
/* Yay, debugging */
int checkForGLErrors(char* s)
{
	int errors = 0 ;

	while ( 1 )
	{
		GLenum x = glGetError() ;

		if ( x == GL_NO_ERROR ){
			if (errors > 0) exit(0);
			return errors ;

		}
		fprintf( stderr, "%s --- OpenGL error: %s [%08x]\n",s, gluErrorString ( x ), x ) ;
		errors++ ;
	}
}
#endif /* DEBUG */

#ifdef LINUX
static SDL_Surface *surface;

/* OpenGL initialisieren, fenster aufmachen, textur malen, etc. */
static void __small__ setup_graphics(void) {

	/* Fenster aufmachen */
	surface = SDL_SetVideoMode( G_WIDTH, G_HEIGHT, G_BPP,SDL_OPENGL|SDL_GL_DOUBLEBUFFER|SDL_HWPALETTE|SDL_HWACCEL|SDL_HWSURFACE);
	if(!surface) {
		my_exit();
	}
	SDL_ShowCursor(0);

	/* OpenGL anschalten */
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);

#ifdef DEBUG
	checkForGLErrors("setup_graphics");
#endif /* DEBUG */
}
#else /* Windows version */
static PIXELFORMATDESCRIPTOR pfd;
static DEVMODE dmScreenSettings;
HDC hDC;
static void setup_graphics(void) {
	dmScreenSettings.dmSize=sizeof(dmScreenSettings);		
	//dmScreenSettings.dmPelsWidth	= G_WIDTH;
	//dmScreenSettings.dmPelsHeight= G_HEIGHT;
	dmScreenSettings.dmBitsPerPel	= 32;
	//dmScreenSettings.dmFields=DM_PELSWIDTH|DM_PELSHEIGHT;
	//ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN);  
	// minimal windows setup code for opengl  
	pfd.cColorBits = pfd.cDepthBits = 32;
	pfd.dwFlags    = PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;	
	// "HDC hDC" changed 19. April 2007 by benny!weltenkonstrukteur.de
	// WS_MAXIMIZE
	hDC = GetDC( CreateWindow("edit", 0, WS_POPUP|WS_VISIBLE, 0, 0, G_WIDTH, G_HEIGHT, 0, 0, 0, 0) );
	SetPixelFormat ( hDC, ChoosePixelFormat ( hDC, &pfd) , &pfd );
	wglMakeCurrent ( hDC, wglCreateContext(hDC) );
}
#endif

/* Functions for creating shaders */
#ifdef DEBUG
static void print_shader_info(unsigned int shader) {
	int len=0;
	char info[1024]; // Yep, this may overflow. Just don't produce that many errors.

	glGetShaderiv(shader, GL_INFO_LOG_LENGTH,&len);
	if(len>0) {
		glGetShaderInfoLog(shader,len,NULL,info);
		fprintf(stderr, "Shader Info Log:\n%s\n",info);
	}
}
#endif /* DEBUG */

static int __small__ create_shader(const char* vertex_source, const char* geometry_source, const char* fragment_source) {
	int vs,fs,gs,prog;

	prog=glCreateProgram();
	vs=glCreateShader(GL_VERTEX_SHADER);
	if(geometry_source) gs=glCreateShader(GL_GEOMETRY_SHADER_EXT);
	else gs=0;
	fs=glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(vs,1,&vertex_source,NULL);
	if(gs) glShaderSource(gs,1,&geometry_source, NULL);
	glShaderSource(fs,1,&fragment_source,NULL);
	glCompileShader(vs);
#ifdef DEBUG
	fprintf(stderr, "Vertex ");
	print_shader_info(vs);
#endif /* DEBUG */
	glCompileShader(fs);
#ifdef DEBUG
	fprintf(stderr, "Fragment ");
	print_shader_info(fs);
	glAttachShader(prog,vs);
	glAttachShader(prog,fs);
#endif /* DEBUG */
	if(gs) {
		glCompileShader(gs);
#ifdef DEBUG
		fprintf(stderr, "Geometry ");
		print_shader_info(gs);
#endif /* DEBUG */
		glAttachShader(prog,gs);

		glProgramParameteriEXT(prog,GL_GEOMETRY_INPUT_TYPE_EXT,GL_POINTS);
		glProgramParameteriEXT(prog,GL_GEOMETRY_OUTPUT_TYPE_EXT,GL_QUADS);
		glGetIntegerv( GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT, &vs );
		glProgramParameteriEXT(prog,GL_GEOMETRY_VERTICES_OUT_EXT,vs);
	}
	glLinkProgram(prog);

	return prog;
}

/* Setup VBO from vertex array */
unsigned int setup_vbo(TexturedVertexData3D* p, size_t n) {
	unsigned int buffer;

	glGenBuffers(1, &buffer);
	glBindBuffer(GL_ARRAY_BUFFER, buffer);
	glBufferData(GL_ARRAY_BUFFER, n, p, GL_STATIC_DRAW);
	return buffer;
}

/* Draw VBO */
void draw_vbo(unsigned int primitive, unsigned int buffer, unsigned int n) {

	/* VBO aktivieren */
	glBindBuffer(GL_ARRAY_BUFFER, buffer);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	/* datentyp, stride, offset, etc. */
	glVertexPointer(3, GL_FLOAT, sizeof(TexturedVertexData3D), 0);
	glNormalPointer(GL_FLOAT, sizeof(TexturedVertexData3D), sizeof(float)*3);
	glTexCoordPointer(2, GL_FLOAT, sizeof(TexturedVertexData3D), sizeof(float)*6);

	/* Do it! */
	glDrawArrays(primitive, 0, n);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
}

/* Copy current framebuffer image to texture */
/* TODO: FBOs stattdessen benutzen */
static void __small__ copy_to_texture(int texture) {
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);
	glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB,0,0,G_WIDTH,G_HEIGHT,0);
}

/* Load Texture from a BMP file */
static int __small__ load_texture(char* filename) {
	int texture;
	unsigned char* pixels; 
	int w,h,n;

#ifndef IMG_LOAD
	/* Without stb_image: BMP-Loader from SDL */
	SDL_Surface* image;
	image = SDL_LoadBMP(filename);

#ifdef DEBUG
	if(!image) {
		fprintf(stderr, "Unable to load texture '%s'!\n",filename);
		return 0;
	}
#endif /* DEBUG */

	pixels=image->pixels;
	w=image->w;
	h=image->h;
	n=3;

#else /* IMG_LOAD */
	/* With stb_image: well, load it. */

	pixels = stbi_load(filename, &w, &h,&n,4);
#ifdef DEBUG
	if(!pixels) {
		fprintf(stderr, "Unable to load texture '%s'!\n",filename);
		return 0;
	}
	n=4;
	fprintf(stderr, "Loaded texture '%s', width %i, height %i\n",filename,w,h);
#endif /* DEBUG */
#endif /* IMG_LOAD */
	glGenTextures(1,&texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexImage2D(GL_TEXTURE_2D, 0, n, w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
#ifdef MIPMAP
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glGenerateMipmapEXT(GL_TEXTURE_2D);
#else
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
#endif /* MIPMAP */


#ifdef IMG_LOAD
	stbi_image_free(pixels);
#endif
	return texture;
}

static int __small__ load_cubemap(char* filename) {
  unsigned int texture;
	unsigned char* pixels; 
	int w,h,n,i;

	pixels = stbi_load(filename, &w, &h,&n,4);
#ifdef DEBUG
	if(!pixels) {
		fprintf(stderr, "Unable to load texture '%s'!\n",filename);
		return 0;
	}
	n=4;
	fprintf(stderr, "Loaded texture '%s', width %i, height %i\n",filename,w,h);
#endif /* DEBUG */

  glGenTextures(1,&texture);
  glBindTexture(GL_TEXTURE_CUBE_MAP , texture);
  for(i=0; i<6;i++){
	  glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, n, w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
  }

  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	stbi_image_free(pixels);
  return texture;
}


/* Turn a quad so that it's facing the camera */
static void __small__ billboard() {
	float modelview[16];
	int i,j;

	/* TODO: das hier kann kuerzer */
	glGetFloatv(GL_MODELVIEW_MATRIX , modelview);

	for(i=0; i<3; i++) {
		for(j=0; j<3; j++) {
			if (i==j)
				modelview[i*4+j] = 1.f;
			else
				modelview[i*4+j] = 0.f;
		}
	}

	glLoadMatrixf(modelview);
}

#ifdef DUMP_VIDEO
void dump_frame(int time) {
	static uint32_t pixels[G_WIDTH * G_HEIGHT];
	char filename[256];
	/* Grab frame into texture */
	glReadPixels(0, 0, G_WIDTH, G_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

	snprintf(filename,255,"frame%06d.bmp",time);
	//fprintf(stderr,"dumping frame %06d\n",time);
	stbi_write_bmp(filename, G_WIDTH, G_HEIGHT, 4, pixels);
}
#endif

/* Flat render (for 2D stuff) */
/* Warning: Push and pop both matrices! */
static void __small__ flat() {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}
